#include <stdio.h>
#include <stdlib.h>
#include <phidget22.h>
#include "PhidgetHelperFunctions.h"

/**
* Configures the device's DataInterval
* Displays info about the attached phidget channel.  
* Fired when a Phidget channel with onAttachHandler registered attaches
*
* @param ph The Phidget channel that fired the attach event
* @param *ctx Context pointer
*/
static void CCONV onAttachHandler(PhidgetHandle ph, void *ctx) {
	PhidgetReturnCode prc;
	int32_t serialNumber;
	PhidgetHandle hub;
	int32_t hubPort;
	int32_t channel;

	/*
	*	Set the DataInterval inside of the attach handler to initialize the device with this value.
	*	DataInterval defines the minimum time between VelocityUpdate events.
	*	DataInterval can be set to any value from MinDataInterval to MaxDataInterval.
	*/
	printf("\tSetting DataInterval to 1000ms\n");
	prc = PhidgetDCMotor_setDataInterval((PhidgetDCMotorHandle)ph, 1000);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Set DataInterval: \n\t");
		DisplayError(prc);
		return;
	}

	prc = Phidget_getDeviceSerialNumber(ph, &serialNumber);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Get DeviceSerialNumber: \n\t");
		DisplayError(prc);
		return;
	}

	prc = Phidget_getChannel(ph, &channel);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Get Channel: \n\t");
		DisplayError(prc);
		return;
	}

	//Check if this is a VINT device
	prc = Phidget_getHub(ph, &hub);
	if (EPHIDGET_WRONGDEVICE != prc) {
		prc = Phidget_getHubPort(ph, &hubPort);
		if (EPHIDGET_OK != prc) {
			fprintf(stderr, "Runtime Error -> Get HubPort: \n\t");
			DisplayError(prc);
			return;
		}
		printf("\n[Attach Event]:\n\t-> Serial Number: %d\n\t-> Hub Port: %d\n\t-> Channel %d\n\n", serialNumber, hubPort, channel);
	}
	else {
		printf("\n[Attach Event]:\n\t-> Serial Number: %d\n\t-> Channel %d\n\n", serialNumber, channel);
	}
}

/**
* Displays info about the detached phidget channel.
* Fired when a Phidget channel with onDetachHandler registered detaches
*
* @param ph The Phidget channel that fired the detach event
* @param *ctx Context pointer
*/
static void CCONV onDetachHandler(PhidgetHandle ph, void *ctx) {
	PhidgetReturnCode prc;
	PhidgetHandle hub;
	int serialNumber;
	int hubPort;
	int channel;

	prc = Phidget_getDeviceSerialNumber(ph, &serialNumber);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Get DeviceSerialNumber: \n\t");
		DisplayError(prc);
		return;
	}

	prc = Phidget_getChannel(ph, &channel);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Get Channel: \n\t");
		DisplayError(prc);
		return;
	}

	//Check if this is a VINT device
	prc = Phidget_getHub(ph, &hub);
	if (EPHIDGET_WRONGDEVICE != prc) {
		prc = Phidget_getHubPort(ph, &hubPort);
		if (EPHIDGET_OK != prc) {
			fprintf(stderr, "Runtime Error -> Get HubPort: \n\t");
			DisplayError(prc);
			return;
		}
		printf("\n[Detach Event]:\n\t-> Serial Number: %d\n\t-> Hub Port: %d\n\t-> Channel %d\n\n", serialNumber, hubPort, channel);
	}
	else {
		printf("\n[Detach Event]:\n\t-> Serial Number: %d\n\t-> Channel %d\n\n", serialNumber, channel);
	}
}

/**
* Writes phidget error info to stderr.
* Fired when a Phidget channel with onErrorHandler registered encounters an error in the library
*
* @param ph The Phidget channel that fired the error event
* @param *ctx Context pointer
* @param errorCode the code associated with the error of enum type Phidget_ErrorEventCode
* @param *errorString string containing the description of the error fired
*/
static void CCONV onErrorHandler(PhidgetHandle phid, void *ctx, Phidget_ErrorEventCode errorCode, const char *errorString) {

	fprintf(stderr, "[Phidget Error Event] -> %s (%d)\n", errorString, errorCode);
}

/**
* Sets the event handlers for Phidget Attach, Phidget Detach, Phidget Error events
*
* @param ph The Phidget channel to add event handlers to
* @return 0 if the operation succeeds, 1 if it fails
*/
int SetAttachDetachError_Handlers(PhidgetHandle ch) {
	PhidgetReturnCode prc;
	printf("\n--------------------------------------\n");
	printf("\nSetting OnAttachHandler...\n");
	prc = Phidget_setOnAttachHandler(ch, onAttachHandler, NULL);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Set Attach Handler: \n\t");
		DisplayError(prc);
		return 1;
	}

	printf("Setting OnDetachHandler...\n");
	prc = Phidget_setOnDetachHandler(ch, onDetachHandler, NULL);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Set Detach Handler: \n\t");
		DisplayError(prc);
		return 1;
	}

	printf("Setting OnErrorHandler...\n");
	prc = Phidget_setOnErrorHandler(ch, onErrorHandler, NULL);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Set Error Handler: \n\t");
		DisplayError(prc);
		return 1;
	}

	return 0;
}

/**
* Outputs the DCMotor's most recently reported velocity.
* Fired when a DCMotor channel with onVelocityUpdateHandler registered has its DataInterval elapse
*
* @param pdch The DCMotor channel that fired the VelocityUpdate event
* @param *ctx Context pointer
* @param velocity The reported velocity from the DCMotor channel
*/
static void CCONV onVelocityUpdateHandler(PhidgetDCMotorHandle pdch, void *ctx, double velocity) {

	printf("[Velocity Event] -> Velocity: %.3g\n", velocity);
}

/**
* Creates a new instance of a DCMotor channel. 
*
* @param pdch Pointer to the PhidgetDCMotorHandle channel to create
* @return 0 if the operation succeeds, 1 if it fails
*/
static int CreateDCMotor(PhidgetDCMotorHandle *pdch) {
	PhidgetReturnCode prc;

	printf("Creating DCMotor Channel...\n");
	prc = PhidgetDCMotor_create(pdch);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Creating DCMotor: \n\t");
		DisplayError(prc);
		return 1;
	}
	return 0;
}


/**
* Sets the TargetVelocity of the DCMotor
*
* @param pdch The PhidgetDCMotorHandle channel to set the TargetVelocity of
* @param velocity The target velocity to send to the DCMotor
* @return 0 if the operation succeeds, 1 if it fails
*/
static int SetDCMotorVelocity(PhidgetDCMotorHandle pdch, double velocity) {
	PhidgetReturnCode prc;

	printf("Setting DCMotor TargetVelocity to %f\n", velocity);
	prc = PhidgetDCMotor_setTargetVelocity(pdch, velocity);
		
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Setting DCMotor Velocity: \n\t");
		DisplayError(prc);
		return 1;
	}
	return 0;
}

/**
* Sets the event handler for DCMotor's VelocityHandler event
*
* @param pdch The PhidgetDCMotorHandle channel to add the event to
* @param fptr The callback function to be called when a VelocityUpdate event is fired 
* @return 0 if the operation succeeds, 1 if it fails
*/
static int SetVelocityHandler(PhidgetDCMotorHandle pdch, PhidgetDCMotor_OnVelocityUpdateCallback fptr) {
	PhidgetReturnCode prc;

	if (fptr) {
		printf("\n--------------------\n"
			"\n  | Velocity updates provide feedback about the current velocity as it accelerates towards the set TargetVelocity\n"
			"  | The linked VelocityUpdate function will run as an event at every DataInterval.\n"
			"  | These events will not occur until a change in velocity >= to the set ChangeTrigger has occured.\n"
			"  | DataInterval and ChangeTrigger should initially be set in the device AttachHandler function.\n");
	}
	printf("\n%s OnVelocityUpdateHandler...\n", (fptr ? "Setting" : "Clearing"));
	printf("\n--------------------\n");

	prc = PhidgetDCMotor_setOnVelocityUpdateHandler(pdch, fptr, NULL);
	if (EPHIDGET_OK != prc) {
		fprintf(stderr, "Runtime Error -> Setting OnVelocityUpdateHandler: \n\t");
		DisplayError(prc);
		return 1;
	}
	return 0;
}


/**
* Creates, configures, and opens a DCMotor channel.
* Provides interface for controlling TargetVelocity of the DCMotor.
* Closes out DCMotor channel when user is finished controlling the device.
*
* @return 0 if the program exits successfully, 1 if it exits with errors.
*/
int main() {
	PhidgetDCMotorHandle ch = NULL;

	/*
	 * Allocate a new Phidget Channel object
	 */

	if (CreateDCMotor(&ch))
		goto error;

	/*
	 * Set matching parameters to specify which channel to open
	 */

	if (SetSerialNumber((PhidgetHandle)ch))
		goto error;

	if (SetVINTProperties((PhidgetHandle)ch))
		goto error;

	if (SetChannel((PhidgetHandle)ch))
		goto error;

	if (SetupNetwork((PhidgetHandle)ch))
		goto error;

	/*
	 * Add event handlers before calling open so that no events are missed.
	 */

	if (SetAttachDetachError_Handlers((PhidgetHandle)ch))
		goto error;

	if (SetVelocityHandler(ch, onVelocityUpdateHandler))
		goto error;

	/*
	 * Open the channel with a timeout
	 */

	if (OpenPhidgetChannel_waitForAttach((PhidgetHandle)ch, 5000))
		goto error;

	
	char buf[100];
	char* endptr;
	int end = 0;
	double velocity = 0.0;

	printf("--------------------\n"
		"\n  | DC motor speed can be controlled by setting its Target Velocity (ie its duty cycle).\n"
		"  | The target velocity can be a number from -1.0 to +1.0, where sign indicates direction of rotation.\n"
		"  | For this example, acceleration has been fixed to 1.0Hz, but can be changed in custom code.\n"

		"\nInput a desired velocity between -1.0 and 1.0 and press ENTER\n"
		"Input Q and press ENTER to quit\n");

	while (!end) {
		if (fgets(buf, 100, stdin) == NULL)
			continue;	

		if (buf[0] == 'Q' || buf[0] == 'q')
			end = 1;
		
		velocity = strtod(buf, &endptr);
		if ((*endptr != '\0') && (isspace(*endptr) == 0))
			continue;

		if (velocity > 1.0 || velocity < -1.0) {
			printf("MAXIMUM velocity is +/-1.0\n");
			continue;
		}
		
		if (SetDCMotorVelocity(ch, velocity))
			goto error;
	}
	
	/*
	 * Perform clean up and exit
	 */

	if (SetVelocityHandler(ch, NULL))
		goto error;

	printf("\nCleaning up...\n");
	ClosePhidgetChannel((PhidgetHandle)ch);
	PhidgetDCMotor_delete(&ch);
	printf("\nExiting...\n");
	printf("Press ENTER to end program.\n");
	getchar();

	return 0;

error:
	PhidgetDCMotor_delete(&ch);
	fprintf(stderr, "\nExiting with error(s)...\n");
	printf("Press ENTER to end program.\n");
	getchar();

	return 1;
}
